"""
    The app tracks the performance of various endpoints over time.
    To bind, use the following lines of code:
        import dashboard
        from flask import Flask
        ...
        app = Flask(__name__)
        ...
        dashboard.bind(app)

    The dashboard with the results that are collected can be found at:
        localhost:5000/dashboard
"""

import os

from flask import Blueprint

from flask_monitoringdashboard.core.config import Config
from flask_monitoringdashboard.core.logger import log

config = Config()


# get current location of the project
def loc():
    return os.path.abspath(os.path.dirname(__file__)) + '/'


# define the blueprint
blueprint = Blueprint('dashboard', __name__, template_folder=loc() + 'templates')


def bind(app, schedule=True):
    """
        Binding the app to this object should happen before importing the routing-
        methods below. Thus, the importing statement is part of this function.
        :param app: the app for which the performance has to be tracked
        :param schedule: flag telling if the background scheduler should be started
    """
    config.app = app
    # Provide a secret-key for using WTF-forms
    if not app.secret_key:
        log('WARNING: You should provide a security key.')
        app.secret_key = 'my-secret-key'

    # Add all route-functions to the blueprint
    from flask_monitoringdashboard.views import deployment, custom, endpoint, outlier, request, profiler, version, auth
    import flask_monitoringdashboard.views

    # Add wrappers to the endpoints that have to be monitored
    from flask_monitoringdashboard.core.measurement import init_measurement
    from flask_monitoringdashboard.core import custom_graph

    blueprint.before_app_first_request(init_measurement)
    if schedule:
        custom_graph.init(app)

    # register the blueprint to the app
    app.register_blueprint(blueprint, url_prefix='/' + config.link)


def add_graph(title, func, **schedule):
    """
    Add a custom graph to the dashboard. You must specify the following arguments
    :param title: title of the graph (must be unique)
    :param schedule: dict containing values for weeks, days, hours, minutes, seconds
    :param func: function reference without arguments
    """
    from flask_monitoringdashboard.core import custom_graph
    graph_id = custom_graph.register_graph(title)
    custom_graph.add_background_job(func, graph_id, **schedule)



from flask_monitoringdashboard import blueprint, config
from flask import redirect, session, render_template, url_for, request
from flask_monitoringdashboard.core.auth import on_logout, on_login

MAIN_PAGE = 'dashboard.index'


@blueprint.route('/login', methods=['GET', 'POST'])
def login():
    """
    User for logging into the system. The POST-request checks whether the logging is valid.
    If this is the case, the user is redirected to the main page.
    :return:
    """
    if session.get(config.link + '_logged_in'):
        return redirect(url_for(MAIN_PAGE))

    if request.method == 'POST':
        name = request.form['name']
        password = request.form['password']
        if name == config.username and password == config.password:
            on_login(admin=True)
            return redirect(url_for(MAIN_PAGE))
        elif name == config.guest_username and password in config.guest_password:
            on_login(admin=False)
            return redirect(url_for(MAIN_PAGE))

    return render_template('fmd_login.html')


@blueprint.route('/logout')
def logout():
    """
    Remove the session variables from the user.
    Redirect the user to the login page.
    :return:
    """
    return on_logout()



from sqlalchemy import func, desc

from flask_monitoringdashboard.database import Request


def get_versions(db_session, endpoint_id=None, limit=None):
    """
    Returns a list of length 'limit' with the versions that are used in the application
    :param db_session: session for the database
    :param endpoint_id: only get the version that are used in this endpoint
    :param limit: only return the most recent versions
    :return: a list of tuples with the versions (as a string) and dates, from oldest to newest
    """
    query = db_session.query(Request.version_requested, func.min(Request.time_requested))
    if endpoint_id:
        query = query.filter(Request.endpoint_id == endpoint_id)
    query = query.group_by(Request.version_requested)
    if limit:
        query = query.limit(limit)
    return query.all()


def get_first_requests(db_session, endpoint_id, limit=None):
    """
    Returns a list with all versions and when they're first used
    :param db_session: session for the database
    :param limit: only return the most recent versions
    :param endpoint_id: id of the endpoint
    :return list of tuples with versions
    """
    query = db_session.query(Request.version_requested, func.min(Request.time_requested).label('first_used')). \
        filter(Request.endpoint_id == endpoint_id). \
        group_by(Request.version_requested).order_by(desc('first_used'))
    if limit:
        query = query.limit(limit)
    return query.all()



"""
    Some useful functions for setting up the testing environment, adding data, etc..
"""
import datetime

from flask import Flask, json

NAME = 'main'
ENDPOINT_ID = 1
REQUEST_IDS = [1, 2, 3, 4, 5]
IP = '127.0.0.1'
GROUP_BY = '1'
REQUESTS = [1000, 2000, 3000, 4000, 50000]
BASE_OUTLIER_EXEC_TIME = 100
TIMES = [datetime.datetime.utcnow()] * 5
OUTLIER_COUNT = 3
for index, _ in enumerate(TIMES):
    TIMES[index] -= datetime.timedelta(seconds=len(REQUESTS) - index)
TEST_NAMES = ['test_name1', 'test_name2']


def set_test_environment():
    """ Override the config-object for a new testing environment. Module flask_monitoringdashboard
    must be imported locally. """
    import flask_monitoringdashboard
    flask_monitoringdashboard.config.database_name = 'sqlite:///test-database.db'


def clear_db():
    """ Drops and creates the tables in the database. Module flask_monitoringdashboard must be imported locally. """
    from flask_monitoringdashboard.database import get_tables, engine
    for table in get_tables():
        table.__table__.drop(engine)
        table.__table__.create(engine)


def add_fake_data():
    """ Adds data to the database for testing purposes. Module flask_monitoringdashboard must be imported locally. """
    from flask_monitoringdashboard.database import session_scope, Request, Endpoint, Outlier
    from flask_monitoringdashboard import config

    # Add requests
    with session_scope() as db_session:
        for i in range(len(REQUESTS)):
            call = Request(id=REQUEST_IDS[i], endpoint_id=ENDPOINT_ID, duration=REQUESTS[i],
                           version_requested=config.version,
                           time_requested=TIMES[i], group_by=GROUP_BY, ip=IP)
            db_session.add(call)

        # Add endpoint
        db_session.add(Endpoint(id=ENDPOINT_ID, name=NAME, monitor_level=1, time_added=datetime.datetime.utcnow(),
                                version_added=config.version, last_requested=TIMES[0]))

        # Add Outliers
        for i in range(OUTLIER_COUNT):
            db_session.add(Outlier(request_id=i+1, cpu_percent='[%d, %d, %d, %d]' % (i, i + 1, i + 2, i + 3)))


def get_test_app(schedule=False):
    """
    :return: Flask Test Application with the right settings
    """
    import flask_monitoringdashboard
    from flask import redirect, url_for
    user_app = Flask(__name__)

    @user_app.route('/')
    def main():
        return redirect(url_for('dashboard.index'))

    @user_app.route('/test_endpoint')
    def test_endpoint():
        return 'endpoint used for testing'

    user_app.config['SECRET_KEY'] = flask_monitoringdashboard.config.security_token
    user_app.testing = True
    flask_monitoringdashboard.user_app = user_app
    user_app.config['WTF_CSRF_ENABLED'] = False
    user_app.config['WTF_CSRF_METHODS'] = []
    flask_monitoringdashboard.config.get_group_by = lambda: '12345'
    flask_monitoringdashboard.bind(app=user_app, schedule=schedule)
    return user_app


def login(test_app):
    """
    Used for setting the sessions, such that you have a fake login to the flask_monitoringdashboard.
    :param test_app:
    """
    from flask_monitoringdashboard import config
    with test_app.session_transaction() as sess:
        sess[config.link + '_logged_in'] = True
        sess[config.link + '_admin'] = True


def test_admin_secure(test_case, page):
    """
    Test whether the page is only accessible with admin credentials.
    :param test_case: test class, must be an instance of unittest.TestCase
    :param page: str with the page of the flask_monitoringdashboard
    """
    with test_case.app.test_client() as c:
        test_case.assertEqual(302, c.get('dashboard/{}'.format(page)).status_code)
        login(c)
        test_case.assertEqual(200, c.get('dashboard/{}'.format(page)).status_code)


def test_get_redirect(test_case, page):
    """
    Test whether the page returns a redirection.
    :param test_case: test class, must be an instance of unittest.TestCase
    :param page: str with the page of the flask_monitoringdashboard
    """
    with test_case.app.test_client() as c:
        test_case.assertEqual(302, c.get('dashboard/{}'.format(page)).status_code)


def test_get_ok(test_case, page):
    """
    Test whether the page returns a 200 OK.
    :param test_case: test class, must be an instance of unittest.TestCase
    :param page: str with the page of the flask_monitoringdashboard
    """
    with test_case.app.test_client() as c:
        test_case.assertEqual(200, c.get('dashboard/{}'.format(page)).status_code)


def test_post_data(test_case, page, data):
    """
    Test whether a post request can successfully be made to the page.
    :param test_case: test class, must be an instance of unittest.TestCase
    :param page: str with the page of the flask_monitoringdashboard
    :param data: the data that should be posted to the page
    """
    with test_case.app.test_client() as c:
        headers = {'content-type': 'application/json'}

        test_case.assertEqual(204, c.post('dashboard/{}'.format(page), data=json.dumps(data),
                                          headers=headers).status_code)



import datetime

from numpy import median

from flask_monitoringdashboard import config
from flask_monitoringdashboard.core.colors import get_color
from flask_monitoringdashboard.core.measurement import add_decorator
from flask_monitoringdashboard.core.timezone import to_local_datetime, to_utc_datetime
from flask_monitoringdashboard.core.utils import simplify
from flask_monitoringdashboard.database import Request
from flask_monitoringdashboard.database.count_group import count_requests_group, get_value
from flask_monitoringdashboard.database.data_grouped import get_endpoint_data_grouped, get_user_data_grouped, \
    get_version_data_grouped
from flask_monitoringdashboard.database.endpoint import get_last_requested, get_endpoints, get_endpoint_by_name, \
    update_endpoint
from flask_monitoringdashboard.database.versions import get_first_requests


def get_endpoint_overview(db_session):
    """
    :param db_session: session for the database
    :return: A list of properties for each endpoint that is found in the database
    """
    week_ago = datetime.datetime.utcnow() - datetime.timedelta(days=7)
    now_local = to_local_datetime(datetime.datetime.utcnow())
    today_local = now_local.replace(hour=0, minute=0, second=0, microsecond=0)
    today_utc = to_utc_datetime(today_local)

    hits_today = count_requests_group(db_session, Request.time_requested > today_utc)
    hits_week = count_requests_group(db_session, Request.time_requested > week_ago)
    hits = count_requests_group(db_session)

    median_today = get_endpoint_data_grouped(db_session, median, Request.time_requested > today_utc)
    median_week = get_endpoint_data_grouped(db_session, median, Request.time_requested > week_ago)
    median_overall = get_endpoint_data_grouped(db_session, median)
    access_times = get_last_requested(db_session)

    return [{
        'id': endpoint.id,
        'name': endpoint.name,
        'monitor': endpoint.monitor_level,
        'color': get_color(endpoint.name),
        'hits-today': get_value(hits_today, endpoint.id),
        'hits-week': get_value(hits_week, endpoint.id),
        'hits-overall': get_value(hits, endpoint.id),
        'median-today': get_value(median_today, endpoint.id),
        'median-week': get_value(median_week, endpoint.id),
        'median-overall': get_value(median_overall, endpoint.id),
        'last-accessed': get_value(access_times, endpoint.name, default=None)
    } for endpoint in get_endpoints(db_session)]


def get_endpoint_users(db_session, endpoint_id, users):
    """
    :param db_session: session for the database
    :param endpoint_id: id for the endpoint
    :param users: a list of users to be filtered on
    :return: a list of dicts with the performance of each user
    """
    times = get_user_data_grouped(db_session, lambda x: simplify(x, 100), Request.endpoint_id == endpoint_id)
    first_requests = get_first_requests(db_session, endpoint_id)
    return [{
        'user': u,
        'date': get_value(first_requests, u),
        'values': get_value(times, u),
        'color': get_color(u)
    } for u in users]


def get_endpoint_versions(db_session, endpoint_id, versions):
    """
    :param db_session: session for the database
    :param endpoint_id: id for the endpoint
    :param versions: a list of version to be filtered on
    :return: a list of dicts with the performance of each version
    """
    times = get_version_data_grouped(db_session, lambda x: simplify(x, 100), Request.endpoint_id == endpoint_id)
    first_requests = get_first_requests(db_session, endpoint_id)
    return [{
        'version': v,
        'date': get_value(first_requests, v),
        'values': get_value(times, v),
        'color': get_color(v)
    } for v in versions]


def get_api_performance(db_session, endpoints):
    """
    :param db_session: session for the database
    :param endpoints: a list of endpoints, encoded by their name
    :return: for every endpoint in endpoints, a list with the performance
    """
    db_endpoints = [get_endpoint_by_name(db_session, end) for end in endpoints]
    data = get_endpoint_data_grouped(db_session, lambda x: simplify(x, 10))
    return [{
        'name': end.name,
        'values': get_value(data, end.id, default=[])
    } for end in db_endpoints]


def set_endpoint_rule(db_session, endpoint_name, monitor_level):
    """
    :param db_session: session for the database
    :param endpoint_name: name of the endpoint
    :param monitor_level: integer, representing the monitoring-level
    :return:
    """
    update_endpoint(db_session, endpoint_name, value=monitor_level)

    # Remove wrapper
    original = getattr(config.app.view_functions[endpoint_name], 'original', None)
    if original:
        config.app.view_functions[endpoint_name] = original
    db_session.commit()

    add_decorator(get_endpoint_by_name(db_session, endpoint_name))



def get_rules(endpoint_name=None):
    """
    :param endpoint_name: if specified, only return the available rules to that endpoint
    :return: A list of the current rules in the attached Flask app
    """
    from flask_monitoringdashboard import config
    try:
        rules = config.app.url_map.iter_rules(endpoint=endpoint_name)
    except KeyError:
        return []
    return [r for r in rules
            if not r.rule.startswith('/' + config.link)
            and not r.rule.startswith('/static-' + config.link)
            and not r.endpoint == 'static']



"""
    This file can be executed for developing purposes.
    It is not used when the flask_monitoring_dashboard is attached to an existing flask application.
"""
import random
import time

from flask import Flask

import flask_monitoringdashboard as dashboard

app = Flask(__name__)

dashboard.config.version = '3.2'
dashboard.config.group_by = '2'
dashboard.config.database_name = 'sqlite:///data.db'
# dashboard.config.database_name = 'mysql+pymysql://user:password@localhost:3306/db1'
dashboard.bind(app)


def f():
    time.sleep(2)
    time.sleep(1)


@app.route('/endpoint')
def endpoint():
    f()
    return 'Ok'


@app.route('/endpoint2')
def endpoint2():
    time.sleep(0.5)
    return 'Ok'


@app.route('/endpoint3')
def endpoint3():
    if random.randint(0, 1) == 0:
        time.sleep(0.1)
    else:
        time.sleep(0.2)
    return 'Ok'


@app.route('/endpoint4')
def endpoint4():
    time.sleep(0.5)
    return 'Ok'


def my_func():
    # here should be something actually useful
    return 33.3


# schedule = {'weeks': 0,
#             'days': 0,
#             'hours': 1,
#             'minutes': 30,
#             'seconds': 0}
#
# dashboard.add_graph('Graph1', lambda: my_func(), **schedule)



import datetime

from flask import jsonify

from flask_monitoringdashboard.controllers.requests import get_num_requests_data, get_hourly_load
from flask_monitoringdashboard.database import session_scope

from flask_monitoringdashboard.core.auth import secure

from flask_monitoringdashboard import blueprint


@blueprint.route('/api/requests/<start_date>/<end_date>')
@secure
def num_requests(start_date, end_date):
    """
    :param start_date: must be in the following form: yyyy-mm-dd
    :param end_date: must be in the following form: yyyy-mm-dd
    :return: A JSON-list with the following object: {
          'name': 'endpoint',
          'values': [list with an integer per day],
        }
    """
    start_date = datetime.datetime.strptime(start_date, '%Y-%m-%d')
    end_date = datetime.datetime.strptime(end_date, '%Y-%m-%d')

    with session_scope() as db_session:
        return jsonify(get_num_requests_data(db_session, start_date, end_date))


@blueprint.route('api/hourly_load/<start_date>/<end_date>')
@blueprint.route('api/hourly_load/<start_date>/<end_date>/<endpoint_id>')
@secure
# both days must be in the form: yyyy-mm-dd
def hourly_load(start_date, end_date, endpoint_id=None):
    """
    :param start_date: must be in the following form: yyyy-mm-dd
    :param end_date: must be in the following form: yyyy-mm-dd
    :param endpoint_id: if specified, filter on this endpoint
    :return: A JSON-object: {
          'data': [ [hits for 0:00-1:00 per day], [hits for 1:00-2:00 per day], ...]
          'days': ['start_date', 'start_date+1', ..., 'end_date'],
        }
    """
    start_date = datetime.datetime.strptime(start_date, '%Y-%m-%d')
    end_date = datetime.datetime.strptime(end_date, '%Y-%m-%d')

    with session_scope() as db_session:
        return jsonify(get_hourly_load(db_session, endpoint_id, start_date, end_date))


"""
    This file contains all unit tests for the endpoint-table in the database. (Corresponding to the file:
    'flask_monitoringdashboard/database/function_calls.py')
    See info_box.py for how to run the test-cases.
"""

import time
import unittest

from flask_monitoringdashboard.database import session_scope
from flask_monitoringdashboard.database.count import count_requests
from flask_monitoringdashboard.database.endpoint import get_endpoint_by_name
from flask_monitoringdashboard.database.request import get_avg_duration
from flask_monitoringdashboard.test.utils import set_test_environment, clear_db, add_fake_data, TIMES, NAME, IP, \
    ENDPOINT_ID


class TestRequest(unittest.TestCase):

    def setUp(self):
        set_test_environment()
        clear_db()
        add_fake_data()

    def test_add_request(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.request import add_request
        name2 = 'main2'
        execution_time = 1234
        self.assertNotEqual(NAME, name2, 'Both cannot be equal, otherwise the test will fail')
        with session_scope() as db_session:
            endpoint = get_endpoint_by_name(db_session, name2)
            self.assertEqual(count_requests(db_session, endpoint.id), 0)
            add_request(db_session, execution_time, endpoint.id, ip=IP)
            self.assertEqual(count_requests(db_session, endpoint.id), 1)


    def test_get_versions(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard import config
        from flask_monitoringdashboard.database.versions import get_versions
        with session_scope() as db_session:
            result = get_versions(db_session)
            self.assertEqual(len(result), 1)
            self.assertEqual(result[0], config.version)

    def test_get_endpoints(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.endpoint import get_endpoints
        with session_scope() as db_session:
            result = get_endpoints(db_session)
            self.assertEqual(len(result), 1)
            self.assertEqual(result[0].name, NAME)

    def test_get_date_of_first_request(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.request import get_date_of_first_request
        with session_scope() as db_session:
            self.assertEqual(get_date_of_first_request(db_session), int(time.mktime(TIMES[0].timetuple())))

    def test_get_avg_duration(self):
        with session_scope() as db_session:
            self.assertEqual(get_avg_duration(db_session, ENDPOINT_ID), 12000)


from __future__ import division
from numpy import std


class GroupedStackLine(object):

    def __init__(self, indent, code, values, total_sum, total_hits):
        self.indent = indent
        self.code = code
        self.values = values
        self.total_sum = total_sum
        self.total_hits = total_hits
        self.index = 0

    @property
    def hits(self):
        return len(self.values)

    @property
    def sum(self):
        return sum(self.values)

    @property
    def standard_deviation(self):
        return std(self.values)

    @property
    def hits_percentage(self):
        return self.hits / self.total_hits

    @property
    def percentage(self):
        return self.sum / self.total_sum

    @property
    def average(self):
        return self.sum / self.hits



import unittest

from flask_monitoringdashboard.test.utils import set_test_environment, clear_db, add_fake_data, get_test_app, \
    test_admin_secure


class TestExportData(unittest.TestCase):

    def setUp(self):
        set_test_environment()
        clear_db()
        add_fake_data()
        self.app = get_test_app()

    def test_download_requests(self):
        """
            Just retrieve the content and check if nothing breaks
        """
        test_admin_secure(self, 'download-requests')

    def test_download_outliers(self):
        """
            Just retrieve the content and check if nothing breaks
        """
        test_admin_secure(self, 'download-outliers')



import unittest

from flask_monitoringdashboard.database import session_scope
from flask_monitoringdashboard.database.count_group import count_requests_group, count_requests_per_day
from flask_monitoringdashboard.test.utils import set_test_environment, clear_db, add_fake_data, TIMES


class TestCountGroup(unittest.TestCase):

    def setUp(self):
        set_test_environment()
        clear_db()
        add_fake_data()

    def test_count_requests_group(self):
        with session_scope() as db_session:
            self.assertEqual(count_requests_group(db_session), [(1, 5)])

    def test_count_requests_per_day(self):
        with session_scope() as db_session:
            self.assertEqual(count_requests_per_day(db_session, []), [])

            if TIMES[0].date() == TIMES[len(TIMES)-1].date():  # must all be on the same day
                self.assertEqual(count_requests_per_day(db_session, [TIMES[0].date()]), [[(1, 5)]])



"""
    Contains all functions that are used to track the performance of the flask-application.
    See init_measurement() for more detailed info.
"""
import time
from functools import wraps

from flask_monitoringdashboard import config
from flask_monitoringdashboard.core.profiler import start_thread_last_requested, start_performance_thread, \
    start_profiler_thread, start_profiler_and_outlier_thread
from flask_monitoringdashboard.core.rules import get_rules
from flask_monitoringdashboard.database import session_scope
from flask_monitoringdashboard.database.endpoint import get_endpoint_by_name


def init_measurement():
    """
    This should be added to the list of functions that are executed before the first request.
    This function is used in the config-method in __init__ of this folder
    It adds wrappers to the endpoints for tracking their performance and last access times.
    """
    with session_scope() as db_session:
        for rule in get_rules():
            endpoint = get_endpoint_by_name(db_session, rule.endpoint)
            add_decorator(endpoint)


def add_decorator(endpoint):
    """
    Adds a wrapper to an endpoint in the app, depending on the monitor level.
    :param endpoint: Endpoint object to be monitored
    """
    fun = config.app.view_functions[endpoint.name]
    if endpoint.monitor_level == 0:
        add_wrapper0(endpoint, fun)
    elif endpoint.monitor_level == 1:
        add_wrapper1(endpoint, fun)
    elif endpoint.monitor_level == 2:
        add_wrapper2(endpoint, fun)
    elif endpoint.monitor_level == 3:
        add_wrapper3(endpoint, fun)
    else:
        raise ValueError('Incorrect monitoringLevel')


def add_wrapper0(endpoint, fun):
    @wraps(fun)
    def wrapper(*args, **kwargs):
        result = fun(*args, **kwargs)
        start_thread_last_requested(endpoint)
        return result

    wrapper.original = fun
    config.app.view_functions[endpoint.name] = wrapper


def add_wrapper1(endpoint, fun):
    @wraps(fun)
    def wrapper(*args, **kwargs):
        start_time = time.time()
        result = fun(*args, **kwargs)
        duration = time.time() - start_time
        start_performance_thread(endpoint, duration)
        return result

    wrapper.original = fun
    config.app.view_functions[endpoint.name] = wrapper


def add_wrapper2(endpoint, fun):
    @wraps(fun)
    def wrapper(*args, **kwargs):
        thread = start_profiler_thread(endpoint)
        start_time = time.time()
        result = fun(*args, **kwargs)
        duration = time.time() - start_time
        thread.stop(duration)
        return result

    wrapper.original = fun
    config.app.view_functions[endpoint.name] = wrapper


def add_wrapper3(endpoint, fun):
    @wraps(fun)
    def wrapper(*args, **kwargs):
        thread = start_profiler_and_outlier_thread(endpoint)
        start_time = time.time()
        result = fun(*args, **kwargs)
        duration = time.time() - start_time
        thread.stop(duration)
        return result

    wrapper.original = fun
    config.app.view_functions[endpoint.name] = wrapper



from collections import defaultdict

import numpy

from flask_monitoringdashboard.core.profiler.util import PathHash
from flask_monitoringdashboard.core.timezone import to_local_datetime
from flask_monitoringdashboard.database import row2dict
from flask_monitoringdashboard.database.stack_line import get_profiled_requests, get_grouped_profiled_requests


def get_profiler_table(db_session, endpoint_id, offset, per_page):
    """
    :param db_session: session for the database
    :param endpoint_id: endpoint to filter on
    :param offset: number of items that are skipped
    :param per_page: number of items that are returned (at most)
    """
    table = get_profiled_requests(db_session, endpoint_id, offset, per_page)

    for idx, row in enumerate(table):
        row.time_requested = to_local_datetime(row.time_requested)
        table[idx] = row2dict(row)
        stack_lines = []
        for line in row.stack_lines:
            obj = row2dict(line)
            obj['code'] = row2dict(line.code)
            stack_lines.append(obj)
        table[idx]['stack_lines'] = stack_lines
    return table


def get_grouped_profiler(db_session, endpoint_id):
    """
    :param db_session: session for the database
    :param endpoint_id: endpoint to filter on
    :return:
    """
    requests = get_grouped_profiled_requests(db_session, endpoint_id)
    db_session.expunge_all()

    histogram = defaultdict(list)  # path -> [list of values]
    path_hash = PathHash()

    for r in requests:
        for index, stack_line in enumerate(r.stack_lines):
            key = path_hash.get_stacklines_path(r.stack_lines, index)
            histogram[key].append(stack_line.duration)

    table = []
    for key, duration_list in sorted(histogram.items(), key=lambda row: row[0]):
        table.append({
            'indent': path_hash.get_indent(key) - 1,
            'code': path_hash.get_code(key),
            'hits': len(duration_list),
            'duration': sum(duration_list),
            'std': numpy.std(duration_list),
            'total_hits': len(requests),
        })
    return table


import datetime

from flask import jsonify

from flask_monitoringdashboard import blueprint, config
from flask_monitoringdashboard.core.auth import secure
from flask_monitoringdashboard.core.timezone import to_local_datetime
from flask_monitoringdashboard.core.utils import get_details
from flask_monitoringdashboard.database import session_scope


@blueprint.route('api/deploy_details')
@secure
def deploy_details():
    """
    :return: A JSON-object with deployment details
    """
    with session_scope() as db_session:
        details = get_details(db_session)
    details['first-request'] = to_local_datetime(datetime.datetime.fromtimestamp(details['first-request']))
    details['first-request-version'] = to_local_datetime(datetime.datetime.
                                                         fromtimestamp(details['first-request-version']))
    return jsonify(details)


@blueprint.route('api/deploy_config')
@secure
def deploy_config():
    """
    :return: A JSON-object with configuration details
    """
    return jsonify({
        'database_name': config.database_name,
        'username': config.username,
        'guest_username': config.guest_username,
        'outlier_detection_constant': config.outlier_detection_constant,
        'timezone': str(config.timezone),
        'colors': config.colors
    })




















"""
    This file contains all unit tests that count a number of results in the database. (Corresponding to the file:
    'flask_monitoringdashboard/database/count.py')
    See info_box.py for how to run the test-cases.
"""

import unittest

from flask_monitoringdashboard.database import session_scope, Endpoint
from flask_monitoringdashboard.database.count import count_rows, count_requests, count_total_requests, \
    count_outliers, count_profiled_requests
from flask_monitoringdashboard.test.utils import set_test_environment, clear_db, add_fake_data, ENDPOINT_ID, REQUESTS, \
    OUTLIER_COUNT


class TestCount(unittest.TestCase):

    def setUp(self):
        set_test_environment()
        clear_db()
        add_fake_data()

    def test_count_rows(self):
        with session_scope() as db_session:
            self.assertEqual(count_rows(db_session, Endpoint.id), 1)
            self.assertEqual(count_rows(db_session, Endpoint.id, Endpoint.id == ENDPOINT_ID+1), 0)

    def test_count_requests(self):
        with session_scope() as db_session:
            self.assertEqual(count_requests(db_session, ENDPOINT_ID), len(REQUESTS))
            self.assertEqual(count_requests(db_session, ENDPOINT_ID + 1), 0)

    def test_count_total_requests(self):
        with session_scope() as db_session:
            self.assertEqual(count_total_requests(db_session), len(REQUESTS))

    def test_count_outliers(self):
        with session_scope() as db_session:
            self.assertEqual(count_outliers(db_session, ENDPOINT_ID), OUTLIER_COUNT)
            self.assertEqual(count_outliers(db_session, ENDPOINT_ID+1), 0)

    def test_count_profiled_requests(self):
        with session_scope() as db_session:
            self.assertEqual(count_profiled_requests(db_session, ENDPOINT_ID), 0)


import atexit
import os

from apscheduler.schedulers import SchedulerAlreadyRunningError
from apscheduler.schedulers.background import BackgroundScheduler

from flask_monitoringdashboard.database import session_scope
from flask_monitoringdashboard.database.custom_graph import add_value, get_graph_id_from_name, get_graphs

scheduler = BackgroundScheduler()


def init(app):
    if not app.debug or os.environ.get('WERKZEUG_RUN_MAIN') == 'true':
        try:
            scheduler.start()
            print('Scheduler started')
            atexit.register(lambda: scheduler.shutdown())
        except SchedulerAlreadyRunningError as err:
            print(err)


def register_graph(name):
    with session_scope() as db_session:
        return get_graph_id_from_name(db_session, name)


def add_background_job(func, graph_id, **schedule):
    def add_data():
        with session_scope() as db_session:
            add_value(db_session, graph_id, func())

    add_data()  # already call once, so it can be verified that the function works
    scheduler.add_job(func=add_data, trigger="interval", **schedule)


def get_custom_graphs():
    with session_scope() as db_session:
        result = get_graphs(db_session)
        db_session.expunge_all()
        return result



import configparser
import unittest


class TestConfig(unittest.TestCase):

    def test_init_from(self):
        """
            Test whether the group_by returns the right result
        """
        import flask_monitoringdashboard as dashboard
        dashboard.config.init_from()
        dashboard.config.init_from(file='../../config.cfg')

    def test_parser(self):
        """
            Test whether the parser reads the right values
        """
        from flask_monitoringdashboard.core.config.parser import parse_literal, parse_bool, parse_string, parse_version
        parser = configparser.RawConfigParser()
        version = '1.2.3'
        string = 'string-value'
        bool = 'False'
        literal = "['a', 'b', 'c']"
        literal2 = '1.23'
        section = 'dashboard'

        parser.add_section(section)
        parser.set(section, 'APP_VERSION', version)
        parser.set(section, 'string', string)
        parser.set(section, 'bool', bool)
        parser.set(section, 'literal', literal)
        parser.set(section, 'literal2', literal2)

        self.assertEqual(parse_version(parser, section, 'default'), version)
        self.assertEqual(parse_string(parser, section, 'string', 'default'), string)
        self.assertEqual(parse_bool(parser, section, 'bool', 'True'), False)
        self.assertEqual(parse_literal(parser, section, 'literal', 'default'), ['a', 'b', 'c'])
        self.assertEqual(parse_literal(parser, section, 'literal2', 'default'), 1.23)

    def test_environment_vars(self):
        """
        Test whether environment variables can be read
        """
        name = 'ENVIRONMENT_VAR'
        value = 'abc'

        from flask_monitoringdashboard.core.config.parser import get_environment_var
        import os
        os.environ[name] = value
        self.assertEqual(get_environment_var(name), value)



"""
    Main class for adding all route-functions to user_app.
    Setup requires only to import this file. All other imports are done in this file
"""
from flask import render_template
from flask.helpers import send_from_directory

from flask_monitoringdashboard import loc, blueprint
from flask_monitoringdashboard.core.auth import secure


@blueprint.route('/static/<path:filename>')
def static(filename):
    """
    Serve static files
    :param filename: filename in the /static file
    :return: content of the file
    """
    return send_from_directory(loc() + 'static', filename)


@blueprint.route('/', defaults={'path': ''})
@blueprint.route('/<path:path>')  # Catch-All URL: http://flask.pocoo.org/snippets/57/
@secure
def index(path):
    return render_template('fmd_base.html')



import unittest

from flask_monitoringdashboard.test.utils import set_test_environment, clear_db, add_fake_data, get_test_app, \
    test_admin_secure


class TestMeasurement(unittest.TestCase):

    def setUp(self):
        set_test_environment()
        clear_db()
        add_fake_data()
        self.app = get_test_app()

    def test_overview(self):
        """
            Just retrieve the content and check if nothing breaks
        """
        test_admin_secure(self, 'overview')

    def test_version_usage(self):
        """
            Just retrieve the content and check if nothing breaks
        """
        test_admin_secure(self, 'version_usage')

    def test_heatmap(self):
        """
            Just retrieve the content and check if nothing breaks
        """
        test_admin_secure(self, 'hourly_load')

    def test_page_number_of_requests_per_endpoint(self):
        """
            Just retrieve the content and check if nothing breaks
        """
        test_admin_secure(self, 'requests')

    def test_page_boxplot_per_endpoint(self):
        """
            Just retrieve the content and check if nothing breaks
        """
        test_admin_secure(self, 'endpoints')



import datetime

import numpy

from flask_monitoringdashboard.core.timezone import to_utc_datetime, to_local_datetime
from flask_monitoringdashboard.database.count_group import count_requests_per_day, get_value
from flask_monitoringdashboard.database.endpoint import get_endpoints, get_num_requests


def get_num_requests_data(db_session, start_date, end_date):
    """
    :param db_session: session for the database
    :param start_date: datetime object
    :param end_date: datetime object and: end_date >= start_date
    :return: a list of the number of requests for each endpoint and on which day
    """
    numdays = (end_date - start_date).days + 1
    days = [start_date + datetime.timedelta(days=i) for i in range(numdays)]

    hits = count_requests_per_day(db_session, days)
    endpoints = get_endpoints(db_session)
    data = [{
        'name': end.name,
        'values': [get_value(hits_day, end.id) for hits_day in hits]
    } for end in endpoints]

    return {
        'days': [d.strftime('%Y-%m-%d') for d in days],
        'data': data
    }


def get_hourly_load(db_session, endpoint_id, start_date, end_date):
    """
    :param db_session: session for the database
    :param endpoint_id: id for the endpoint
    :param start_date: datetime object
    :param end_date: datetime object and: end_date >= start_date
    :return:
    """
    numdays = (end_date - start_date).days + 1

    # list of hours: 0:00 - 23:00
    hours = ['0{}:00'.format(h) for h in range(0, 10)] + ['{}:00'.format(h) for h in range(10, 24)]
    heatmap_data = numpy.zeros((len(hours), numdays))

    start_datetime = to_utc_datetime(datetime.datetime.combine(start_date, datetime.time(0, 0, 0, 0)))
    end_datetime = to_utc_datetime(datetime.datetime.combine(end_date, datetime.time(23, 59, 59)))

    for time, count in get_num_requests(db_session, endpoint_id, start_datetime, end_datetime):
        parsed_time = datetime.datetime.strptime(time, '%Y-%m-%d %H:%M:%S')
        day_index = (parsed_time - start_datetime).days
        hour_index = int(to_local_datetime(parsed_time).strftime('%H'))
        heatmap_data[hour_index][day_index] = count
    return {
        'days': [(start_date + datetime.timedelta(days=i)).strftime('%Y-%m-%d') for i in range(numdays)],
        "data": heatmap_data.tolist()
    }


"""
    This file contains all unit tests for the monitor-rules-table in the database. (Corresponding to the file:
    'flask_monitoringdashboard/database/outlier.py')
    See info_box.py for how to run the test-cases.
"""

import unittest

from flask_monitoringdashboard.database import session_scope
from flask_monitoringdashboard.database.outlier import add_outlier
from flask_monitoringdashboard.test.utils import set_test_environment, clear_db, add_fake_data, NAME, OUTLIER_COUNT, \
    ENDPOINT_ID


class TestMonitorRule(unittest.TestCase):

    def setUp(self):
        set_test_environment()
        clear_db()
        add_fake_data()

    def test_add_outlier(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.outlier import Outlier
        with session_scope() as db_session:
            self.assertEqual(len(db_session.query(Outlier).all()), OUTLIER_COUNT)
            request = "headers", "environ", "url"
            add_outlier(db_session, request_id=1, cpu_percent="cpu_percent", memory="memory",
                        stacktrace="stacktrace", request=request)
            self.assertEqual(len(db_session.query(Outlier).all()), OUTLIER_COUNT+1)

    def test_get_outliers(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.outlier import get_outliers_sorted
        with session_scope() as db_session:
            outliers = get_outliers_sorted(db_session, endpoint_id=ENDPOINT_ID, offset=0, per_page=10)
        self.assertEqual(len(outliers), OUTLIER_COUNT)
        for i, outlier in enumerate(outliers):
            self.assertEqual(outlier.request.endpoint.name, NAME)
            if i == 0:
                continue
            self.assertTrue(outlier.request.time_requested <= outliers[i - 1].request.time_requested)

    def test_count_outliers(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.count import count_outliers
        with session_scope() as db_session:
            self.assertEqual(count_outliers(db_session, ENDPOINT_ID), OUTLIER_COUNT)

    def test_get_outliers_cpus(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.outlier import get_outliers_cpus
        expected_cpus = []
        for i in range(OUTLIER_COUNT):
            expected_cpus.append('[%d, %d, %d, %d]' % (i, i + 1, i + 2, i + 3))
        with session_scope() as db_session:
            self.assertEqual(get_outliers_cpus(db_session, ENDPOINT_ID), expected_cpus)



import threading

from flask import request

from flask_monitoringdashboard.core.profiler.baseProfiler import BaseProfiler
from flask_monitoringdashboard.core.profiler.outlierProfiler import OutlierProfiler
from flask_monitoringdashboard.core.profiler.performanceProfiler import PerformanceProfiler
from flask_monitoringdashboard.core.profiler.stacktraceProfiler import StacktraceProfiler


def start_thread_last_requested(endpoint):
    """
    Starts a thread that updates the last_requested time in the database.
    :param endpoint: Endpoint object
    """
    BaseProfiler(endpoint).start()


def start_performance_thread(endpoint, duration):
    """
    Starts a thread that updates performance, utilization and last_requested in the database.
    :param endpoint: Endpoint object
    :param duration: duration of the request
    """
    ip = request.environ['REMOTE_ADDR']
    PerformanceProfiler(endpoint, ip, duration).start()


def start_profiler_thread(endpoint):
    """ Starts a thread that monitors the main thread. """
    current_thread = threading.current_thread().ident
    ip = request.environ['REMOTE_ADDR']
    thread = StacktraceProfiler(current_thread, endpoint, ip)
    thread.start()
    return thread


def start_profiler_and_outlier_thread(endpoint):
    """ Starts two threads: PerformanceProfiler and StacktraceProfiler.  """
    current_thread = threading.current_thread().ident
    ip = request.environ['REMOTE_ADDR']
    outlier = OutlierProfiler(current_thread, endpoint)
    thread = StacktraceProfiler(current_thread, endpoint, ip, outlier)
    thread.start()
    outlier.start()
    return thread



"""
Contains all functions that access an StackLine object.
"""

from sqlalchemy import desc, distinct
from sqlalchemy.orm import joinedload

from flask_monitoringdashboard.database import StackLine, Request
from flask_monitoringdashboard.database.code_line import get_code_line


def add_stack_line(db_session, request_id, position, indent, duration, code_line):
    """
    Adds a StackLine to the database (and possibly a CodeLine)
    :param db_session: Session for the database
    :param request_id: id of the request
    :param position: position of the StackLine
    :param indent: indent-value
    :param duration: duration of this line (in ms)
    :param code_line: quadruple that consists of: (filename, line_number, function_name, code)
    """
    fn, ln, name, code = code_line
    db_code_line = get_code_line(db_session, fn, ln, name, code)
    db_session.add(StackLine(request_id=request_id, position=position, indent=indent, code_id=db_code_line.id,
                             duration=duration))


def get_profiled_requests(db_session, endpoint_id, offset, per_page):
    """
    Gets the requests of an endpoint sorted by request time, together with the stack lines.
    :param db_session: session for the database
    :param endpoint_id: filter profiled requests on this endpoint
    :param offset: number of items to skip
    :param per_page: number of items to return
    :return: A list with tuples. Each tuple consists first of a Request-object, and the second part of the tuple
            is a list of StackLine-objects.
    """
    result = db_session.query(Request).filter(Request.endpoint_id == endpoint_id). \
        options(joinedload(Request.stack_lines).joinedload(StackLine.code)). \
        filter(Request.stack_lines.any()). \
        order_by(desc(Request.time_requested)). \
        offset(offset).limit(per_page).all()
    db_session.expunge_all()
    return result


def get_grouped_profiled_requests(db_session, endpoint_id):
    """
    Gets the grouped stack lines of all requests of an endpoint.
    :param db_session: session for the database
    :param endpoint_id: filter profiled requests on this endpoint
    :return: A list with tuples. Each tuple consists first of a Request-object, and the second part of the tuple
            is a list of StackLine-objects.
    """
    t = db_session.query(distinct(StackLine.request_id).label('id')). \
        filter(Request.endpoint_id == endpoint_id). \
        join(Request.stack_lines).order_by(StackLine.request_id.desc()).limit(100).subquery('t')
    # Limit the number of results by 100, otherwise the profiler gets too large
    # and the page doesn't load anymore. We show the most recent 100 requests.
    result = db_session.query(Request). \
        join(Request.stack_lines). \
        filter(Request.id == t.c.id). \
        order_by(desc(Request.id)). \
        options(joinedload(Request.stack_lines).joinedload(StackLine.code)).all()
    db_session.expunge_all()
    return result



from colorhash import ColorHash
from flask_monitoringdashboard import config
import re


def get_color(hash):
    """
    Returns an rgb-color (as string, which can be using in plotly) from a given hash,
    if no color for that string was specified in the config file.
    :param hash: the string that is translated into a color
    :return: a color (as string)
    """
    if hash in config.colors:
        rgb = re.findall(r'\d+', config.colors[hash])
    else:
        rgb = ColorHash(hash).rgb
    return 'rgb({0}, {1}, {2})'.format(rgb[0], rgb[1], rgb[2])



import json
import os

import setuptools

loc = os.path.dirname(os.path.abspath(__file__))


def get_description():
    with open(loc + '/README.md') as readme:
        info = readme.read()
    with open(loc + '/docs/changelog.rst') as changelog:
        return info + '\n\n' + changelog.read()


with open(loc + '/requirements.txt') as f:
    required = f.read().splitlines()

with open('flask_monitoringdashboard/constants.json', 'r') as f:
    constants = json.load(f)

setuptools.setup(
    name="Flask-MonitoringDashboard",
    version=constants['version'],
    packages=setuptools.find_packages(),
    include_package_data=True,
    platforms='Any',
    zip_safe=False,
    test_suite='flask_monitoringdashboard.test.get_test_suite',
    url='https://github.com/flask-dashboard/Flask-MonitoringDashboard',
    author=constants['author'],
    author_email=constants['email'],
    description="Automatically monitor the evolving performance of Flask/Python web services.",
    long_description=get_description(),
    install_requires=required,
    entry_points={
        'flask.commands': [
            'fmd=flask_monitoringdashboard.cli:fmd'
        ],
    },
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Framework :: Flask'],
    project_urls={
        'Bug Reports': 'https://github.com/flask-dashboard/Flask-MonitoringDashboard/issues',
        'PyPi': 'https://pypi.org/project/Flask-MonitoringDashboard/',
        'Documentation': 'http://flask-monitoringdashboard.readthedocs.io/',
        'Source': 'https://github.com/flask-dashboard/Flask-MonitoringDashboard/',
    },
)



import unittest

from flask_monitoringdashboard.test.utils import set_test_environment, clear_db, add_fake_data, \
    get_test_app, login, NAME, test_admin_secure, test_get_ok, test_get_redirect


class TestSetup(unittest.TestCase):

    def setUp(self):
        set_test_environment()
        clear_db()
        add_fake_data()
        self.app = get_test_app()

    def test_index(self):
        """
            Just retrieve the content and check if nothing breaks
        """
        test_get_redirect(self, '')

    def test_static(self):
        """
            Just retrieve the content and check if nothing breaks
        """
        test_get_ok(self, 'static/css/custom.css')

    def test_configuration(self):
        """
            Just retrieve the content and check if nothing breaks
        """
        test_admin_secure(self, 'configuration')

    def test_rules(self):
        """
            Just retrieve the content and check if nothing breaks
        """
        test_admin_secure(self, 'rules')

    def test_monitor_rule(self):
        """
            Test whether it is possible to monitor a rule
        """
        data = {'name': NAME, 'value': 0}
        with self.app.test_client() as c:
            login(c)
            self.assertEqual(200, c.post('dashboard/api/set_rule', data=data).status_code)
